//if #define TextMode will show all messages for debug
//#define TextMode

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "bi_io.h"
#include "hufftab.h"

#define MASK  0x7FFFFFFF

/* IMPORTANT TEST for signed or unsigned int shift:
    
  (1) TEST 1
    unsigned int test=0xF0000000;
    test >> 4 will result 0x0F000000

    int test=0xF0000000;
    test >> 4 will result 0xFF000000

    For shifting left, the bits will be filled with 0. 

  (2) TEST 2
    test << (n*32+i) = test << i
    test >> (n*32+i) = test >> i
*/

/* Get bits from a opened file:
    current_word : save the current bit stream
                   the word is always full.
    temp_word : it is a partial valid word in the order byte1,2,3,4
    index : show how many bits is valid in the temp_word
     (ex) temp_word:
          |0000 0000|0000 0000|0000 0000|00?? ????|
                                           ^index=6 (from 1 to 32)
*/
void build_table(table_elem TAB[], int max_len, int *tab);
void print_table(int *table, int max_len);
void print_dct_table(int *table, int max_len);

bi_io_class::bi_io_class(FILE* _in){
    unsigned int buffer;
    in=_in;
    index=32;
    err=stdout;
    if(in==NULL){ 
        fprintf(err,"ERROR: file does not exist!\n");
        exit(0);
    }
    flag= &(in->_flag);
    buffer = (unsigned int) getw(in);
    if((*flag) & _IOEOF) current_word=0x000001B7;
    else{
        current_word = buffer << 24;
        current_word |= ( (buffer<<8) & 0x00ff0000 );
        current_word |= ( (buffer>>8) & 0x0000ff00 );
        current_word |= ( (buffer>>24)& 0x000000ff );
    }
    buffer = (unsigned int) getw(in);
    if((*flag) & _IOEOF) temp_word=0x000001B7;
    else{
        temp_word = buffer << 24;
        temp_word |= ( (buffer<<8) & 0x00ff0000 );
        temp_word |= ( (buffer>>8) & 0x0000ff00 );
        temp_word |= ( (buffer>>24)& 0x000000ff );
    }
    mb_increment_table = new int[2048];
    memset(mb_increment_table,0,sizeof(int[2048]));
    build_table(AddrInc,11,mb_increment_table);
    #ifdef TextMode
        printf("\nmb_increment_table: \n");
        print_table(mb_increment_table,11);
    #endif

    build_mb_type_p_table();
    #ifdef TextMode
        printf("\nmb_type_p_table: \n");
        print_table(mb_type_p_table, 6);
    #endif

    build_mb_type_b_table();
    #ifdef TextMode
        printf("\nmb_type_b_table:\n");
        print_table(mb_type_b_table, 6);
    #endif

    mb_pattern_table = new int[512];
    memset(mb_pattern_table,0,sizeof(int[512]));
    build_table(MacroPatt,9,mb_pattern_table);
    #ifdef TextMode
        printf("\nmb_pattern_table:\n");
        print_table(mb_pattern_table, 9);
    #endif

    mv_table = new int[2048];
    memset(mv_table,0,sizeof(int[2048]));
    build_table(MotionVector,11,mv_table);
    #ifdef TextMode
        printf("\nmv_table:\n");
        print_table(mv_table, 11);
    #endif

    dc_l_table = new int[128];
    memset(dc_l_table,0,sizeof(int[128]));
    build_table(DCTDCSizeL,7,dc_l_table);
    #ifdef TextMode
        printf("\ndc_l_table:\n");
        print_table(dc_l_table,7);
    #endif

    dc_c_table = new int[256];
    memset(dc_c_table,0,sizeof(int[256]));
    build_table(DCTDCSizeC,8,dc_c_table);
    #ifdef TextMode
        printf("\ndc_c_table:\n");
        print_table(dc_c_table,8);
    #endif

    build_dct_table();
    #ifdef TextMode
        printf("\ndct_1st_table:\n");
        print_dct_table(dct_1st_table, 12);
        printf("\ndct_2nd_table:\n");
        print_dct_table(dct_2nd_table, 8);
    #endif
}

bi_io_class::~bi_io_class(void){ 
    if(mb_increment_table!=NULL) delete mb_increment_table;
    if(mb_type_p_table!=NULL) delete mb_type_p_table;
    if(mb_type_b_table!=NULL) delete mb_type_b_table;
    if(mb_pattern_table!=NULL) delete mb_pattern_table;
    if(mv_table!=NULL) delete mv_table;
    if(dc_l_table!=NULL) delete dc_l_table;
    if(dc_c_table!=NULL) delete dc_c_table;
    if(dct_1st_table!=NULL) delete dct_1st_table;
    if(dct_2nd_table!=NULL) delete dct_2nd_table;
}
    
unsigned int bi_io_class::getn(int n){
    unsigned int result;
    //printf("getn(%i) [%X %8X index=%i]\n",n,current_word,temp_word,index);
    if(n>32 || n<0){
        fprintf(err,"ERROR: getn(%i)\n", n);
    }
    if(n==0) return 0;
    if(n==32){
        result=current_word;
        current_word=0;
    }else{
        result=current_word>>(32-n);
        current_word <<= n;
    }
    index-=n;
    if( index > 0 ){
        current_word |= (temp_word >> index);
        temp_word &= (MASK >> (31-index));
    }else{
        current_word |= (temp_word << (-index));
        unsigned int buffer;
        index+=32;
        buffer = (unsigned int) getw(in);
        if((*flag) & _IOEOF) temp_word=0x000001B7;
        else{
            temp_word = buffer << 24;
            temp_word |= ( (buffer<<8) & 0x00ff0000 );
            temp_word |= ( (buffer>>8) & 0x0000ff00 );
            temp_word |= ( (buffer>>24)& 0x000000ff );
        }
        if(index==32) return result;
        else{
            current_word |= (temp_word >> index);
            temp_word &= (MASK >> (31-index));
        }
    }
    return result;
}

unsigned int bi_io_class::nextn(int n){
    return (current_word >> (32-n));
}

unsigned int bi_io_class::next24(void){
    return (current_word >> 8);
}

unsigned int bi_io_class::next23(void){
    return (current_word >> 9);
}

unsigned int bi_io_class::next1(void){
    return (current_word >> 31);
}

unsigned int bi_io_class::next32(void){
    return current_word;
}

int bi_io_class::eof(void){
    return (*flag)&_IOEOF;
}

int bi_io_class::bytealigned(void){
    return !(index%8) ;
}

void bi_io_class::next_start_code(void){
    while(index%8) getn(1);
    while(next24()!=0x00000001){
        getn(8);
    }
}

/**********************************************************
    Table initialization

    Each entry is 
        byte1       byte2      byte3       byte4
    | ........value........ |0000 0000 | codelength | 

    Because each entry is int type, if value is negtive,
    '>>' is a signed shift operation.

***********************************************************/
void build_table(table_elem TAB[], int max_len, int *tab){
    int i;
    table_elem elem;
    elem = TAB[0];
    for(i=0; elem.code[0]!=0; i++, elem=TAB[i]){
        int length, code_value, value, times, j;
        char temp;
        code_value=0;
        length=0;
        temp=elem.code[0];
        for(j=0; temp!=0; j++, temp=elem.code[j])
            if(temp=='1' || temp=='0'){
                code_value = code_value*2 + temp - '0';
                length++;
            }
        value= (elem.value<<16) | length;
        times = 1 << (max_len-length);
        code_value <<= (max_len-length);
        for(j=0; j<times; j++)
            tab[code_value+j]=value;
    }
}

void bi_io_class::build_mb_type_p_table(void){
    int i;
    mb_type_p_table = new int[64];
    memset(mb_type_p_table,0,sizeof(int[64]));
    for(i=0; MacroTypeP[i].code[0]!=0; i++){
        int length, code_value, value, times, j;
        char temp;
        mb_type_elem mb_type;
        mb_type=MacroTypeP[i];
        code_value=0;
        length=0;
        temp=mb_type.code[0];
        for(j=0; temp!=0; j++, temp=mb_type.code[j])
            if(temp!=' '){
                code_value = code_value*2 + temp - '0';
                length++;
            }
        value=(mb_type.mb_q<<4)|(mb_type.mb_mf<<3)|(mb_type.mb_mb<<2)
               |(mb_type.mb_p<<1) | (mb_type.mb_i);
        value=(value<<16) | length;
        times = 1 << (6-length);
        code_value <<= (6-length);
        for(j=0; j<times; j++)
            mb_type_p_table[code_value+j]=value;
    }
}

void bi_io_class::build_mb_type_b_table(void){
    int i;
    mb_type_b_table = new int[64];
    memset(mb_type_b_table,0,sizeof(int[64]));
    for(i=0; MacroTypeB[i].code[0]!=0; i++){
        int length, code_value, value, times, j;
        char temp;
        mb_type_elem mb_type;
        mb_type=MacroTypeB[i];
        code_value=0;
        length=0;
        temp=mb_type.code[0];
        for(j=0; temp!=0; j++, temp=mb_type.code[j])
            if(temp!=' '){
                code_value = code_value*2 + temp - '0';
                length++;
            }
        value=(mb_type.mb_q<<4)|(mb_type.mb_mf<<3)|(mb_type.mb_mb<<2)
               |(mb_type.mb_p<<1) | (mb_type.mb_i);
        value=(value<<16) | length;
        times = 1 << (6-length);
        code_value <<= (6-length);
        for(j=0; j<times; j++)
            mb_type_b_table[code_value+j]=value;
    }
}

/* DCT Coefficient look-up Table

      byte1       byte2        byte3       byte4
  |    run    |   level    | 0000 0000 | codelength |

*/
void bi_io_class::build_dct_table(void){
    int i;
    dct_elem elem;
    dct_1st_table = new int[4096];
    memset(dct_1st_table,0,sizeof(int[4096]));
    elem = DCTCoeff[0];
    for(i=0;; i++, elem=DCTCoeff[i]){
        int length, code_value, value, times, j;
        char temp;
        code_value=0;
        length=0;
        temp=elem.code[0];
        for(j=0; temp!=0; j++, temp=elem.code[j])
            if(temp=='1' || temp=='0'){
                code_value = code_value*2 + temp - '0';
                length++;
            }
        if(length>12) break; // 1st table is finished.
        value= (elem.run<<24) | (elem.level<<16) | length;
        times = 1 << (12-length);
        code_value <<= (12-length);
        for(j=0; j<times; j++)
            dct_1st_table[code_value+j]=value;
    }
    dct_2nd_table = new int[256];
    memset(dct_2nd_table, 0, sizeof(int[256]));
    for(; elem.code[0]!=0; i++, elem=DCTCoeff[i]){
        int length, code_value, value, times, j;
        char temp;
        code_value=0;
        length=0;
        temp=elem.code[0];
        for(j=0; temp!=0; j++, temp=elem.code[j])
            if(temp=='1' || temp=='0'){
                code_value = code_value*2 + temp - '0';
                length++;
            }
        length -= 8;  // the first 8 bits are all zero
        value= (elem.run<<24) | (elem.level<<16) | length;
        times = 1 << (8-length);
        code_value <<= (8-length);
        for(j=0; j<times; j++)
            dct_2nd_table[code_value+j]=value;
    }
}

/**********************************************************
    Table lookup
***********************************************************/

inline int bi_io_class::table_lookup(int table[], int max_len){
    int result, n;
    result=table[current_word>>(32-max_len)];
    if(result==0){
        fprintf(err,"ERROR: access undefine codeword!\n");
    }
    n=result&0xFF;
    if(n>32 || n<0){
        fprintf(err,"ERROR: codelength=%i\n", n);
    }
    //printf("table_lookup() [%X %8X index=%i] length=%i max_len=%i\n",
    //    current_word,temp_word,index,n,max_len);
    result >>= 16;
    current_word <<= n;
    index-=n;
    if( index > 0 ){
        current_word |= (temp_word >> index);
        temp_word &= (MASK >> (31-index));
    }else{
        current_word |= (temp_word << (-index));
        unsigned int buffer;
        index+=32;
        buffer = (unsigned int) getw(in);
        if((*flag) & _IOEOF) temp_word=0x000001B7;
        else{
            temp_word = buffer << 24;
            temp_word |= ( (buffer<<8) & 0x00ff0000 );
            temp_word |= ( (buffer>>8) & 0x0000ff00 );
            temp_word |= ( (buffer>>24)& 0x000000ff );
        }
        if(index==32) return result;
        else{
            current_word |= (temp_word >> index);
            temp_word &= (MASK >> (31-index));
        }
    }
    return result;
}

int bi_io_class::get_mb_address_increment(void){
    int result;
    result=0;
    while( (current_word>>21) == 0x0F ) getn(11);
    for(;;){
        result+=table_lookup(mb_increment_table,11);
        if(result!=35) break;  //35 denote macroblock_escape
        else result-=2;        //  the value is 33 ,so minus 2
    }
    return result;
}

int bi_io_class::get_mb_type(int i){ //i denote picture type
    switch(i){
    case 1: // for I-picture
        if(getn(1)==1)
            return 0x01; //0,0,0,0,1
        else{
            if(getn(1)!=1)
                fprintf(err,"ERROR: get_mb_type (1)\n");
            return 0x11; //1,0,0,0,1
        }
    case 2: // for P-picture
        return table_lookup(mb_type_p_table,6);
    case 3: // for B-picutre
        return table_lookup(mb_type_b_table,6);
    case 4: // for D-picture
        getn(1);
        return 0x01;
    }
    return 0;
}

int bi_io_class::get_cbp(void){
    return table_lookup(mb_pattern_table,9);
}

int bi_io_class::get_mv(void){
    return table_lookup(mv_table,11);
}

int bi_io_class::get_dc_size_l(void){
    return table_lookup(dc_l_table,7);
}

int bi_io_class::get_dc_size_c(void){
    return table_lookup(dc_c_table,8);
}

void bi_io_class::get_dct_first(int& run, int& level){
    if(next1()==1){
        getn(1);
        if(getn(1)==0){
            run=0; level=1;
        }else{
            run=0; level=-1;
        }
        return;
    }
    if(current_word & 0xFF000000){ //at least one of first 8 is 1.
        run=table_lookup(dct_1st_table,12);
        level = run&0x00FF;
        run >>= 8;
        if(run==64){
            run=(int)getn(6);
            level=(int)getn(8);
            if(level==0)
                level=(int)getn(8);
            else if(level==128)
                level=(int)getn(8)-256;
            else if(level&0x80) // negtive
                level-=256;
            return;
        }
    }else{
        getn(8);
        run=table_lookup(dct_2nd_table,8);
        level = run&0x00FF;
        run >>= 8;
    }
    if(getn(1)==1)
        level = -level;
}

void bi_io_class::get_dct_next(int& run, int& level){
    //printf("get_dct() [%X %8X index=%i]\n",
    //    current_word,temp_word,index);
    if((current_word>>30)==2){
        getn(2);
        run=0; level=0; return;
    }
    if(current_word & 0xFF000000){ //at least one of first 8 is 1.
        run=table_lookup(dct_1st_table,12);
        level = run&0x00FF;
        run >>= 8;
        if(run==64){
            run=(int)getn(6);
            level=(int)getn(8);
            if(level==0)
                level=(int)getn(8);
            else if(level==128)
                level=(int)getn(8)-256;
            else if(level&0x80) // negtive
                level-=256;
            return;
        }
    }else{
        getn(8);
        run=table_lookup(dct_2nd_table,8);
        level = run&0x00FF;
        run >>= 8;
    }
    if(getn(1)==1)
        level = -level;
}

#ifdef TextMode
void print_table(int* table, int max_len){
    int i, size, value, length, j, code, times, prev_elem;
    size = 1 << max_len;
    for(i=0; i<size; ){
        if(table[i]==0){ i++; continue;}
        prev_elem=table[i];
        value=table[i] >> 16;
        length=table[i] & 0xFF;
        code=i >> (max_len-length);
        for(j=length; j>0; j--)
            printf("%i", (code >> (j-1))&0x01);
        for(j=0; j <= max_len-length; j++)
            printf(" ");
        printf("    value=%i length=%i\n", value, length);
        times= 1 << (max_len-length);
        for(j=0;j<times; j++, i++)
            if(table[i] != prev_elem)
                fprintf(err,"ERROR: table entry error\n");
    }
}

void print_dct_table(int* table, int max_len){
    int i, size, value, length, j, code, times, prev_elem;
    size = 1 << max_len;
    for(i=0; i<size; ){
        if(table[i]==0){ i++; continue;}
        prev_elem=table[i];
        value=table[i] >> 16;
        length=table[i] & 0xFF;
        code=i >> (max_len-length);
        for(j=length; j>0; j--)
            printf("%i", (code >> (j-1))&0x01);
        for(j=0; j <= max_len-length; j++)
            printf(" ");
        printf("    run=%i level=%i length=%i\n", 
            value>>8, value&0xFF ,length);
        times= 1 << (max_len-length);
        for(j=0;j<times; j++, i++)
            if(table[i] != prev_elem)
                fprintf(err,"ERROR: table entry error\n");
    }
}
#endif

//the following is the previous implementation!!
/* Get bits from a opened file:
    current_word : save the current bit stream
    index : show how many bits is valid in the current_word
     (ex) current_word:
          |0000 0000|0000 0000|0000 0000|00?? ????|
                                           ^index=6 (from 0 to 32)
    temp_word : is always full, in the order byte1,2,3,4
*/
/*
bi_io_class::bi_io_class(FILE* in){
    unsigned int buffer;
    index=0;
    current_word=0;
    temp_word=0;
    if(in==NULL){ 
        fprintf(err,"ERROR: file does not exist!\n");
        exit(0);
    }
    flag= &(in->_flag);
    buffer = (unsigned int) getw(in);
    temp_word = buffer << 24;
    temp_word |= ( (buffer<<8) & 0x00ff0000 );
    temp_word |= ( (buffer>>8) & 0x0000ff00 );
    temp_word |= ( (buffer>>24)& 0x000000ff );
}

bi_io_class::~bi_io_class(void){ }
    
unsigned int bi_io_class::getn(int n){
    unsigned int result;
    if(n==0) return 0;
    index-=n;
    if( index >= 0 ){
        result = current_word >> index;
        current_word &= MASK >> (31-index);
    }else{
        unsigned int buffer;
        result = current_word << (-index);
        index+=32;
        current_word=temp_word;
        buffer = (unsigned int) getw(in);
        temp_word = buffer << 24;
        temp_word |= ( (buffer<<8) & 0x00ff0000 );
        temp_word |= ( (buffer>>8) & 0x0000ff00 );
        temp_word |= ( (buffer>>24)& 0x000000ff );
        result |= current_word >> index;
        current_word &= MASK >> (31-index);
    }
    return result;
}

unsigned int bi_io_class::nextn(int n){
    int temp;
    temp=index;
    if(n==0) return 0;
    temp-=n;
    if( temp >= 0 )
        return current_word >> temp;
    else
        return (current_word << (-temp)) | (temp_word >> (temp+32));
}

int bi_io_class::eof(void){
    return (*flag)&_IOEOF;
}

int bi_io_class::bytealigned(void){
    return !(index%8) ;
}
*/

