/*
* MPEG player
*/

#define IDS_TEST      3000
#define IDO_TEST      4000

#define WINDOW_WIDTH  500
#define WINDOW_HEIGHT 400

#include <windows.h>
#include <ddraw.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <process.h>
#include <time.h>
#include "mpeg.h"

OPENFILENAME ofile;
char file[256];
int Ximage, Yimage;

int* IMAGE;

extern void Image_YCbCr_to_BGR(int* YCbCrImage, char* RGBImage, 
                               int h_size, int v_size);

//char* IMGTMP[2];
//#define WINDOW_WIDTH  500
//#define WINDOW_HEIGHT 400

LRESULT WINAPI ClientWndProc(HWND,UINT,WPARAM,LPARAM);

//for DirectDraw
char*               Start;
LONG                NextScan;
RECT                rcWindow;
IDirectDraw         *DDraw;
IDirectDrawSurface  *FrontBuffer;
IDirectDrawSurface  *BackBuffer;
IDirectDrawClipper  *Clipper;

char buffer1[80];
char buffer2[80];


HWND winhwnd;

BOOL InitDDraw(HWND hWnd, int h_size, int v_size)
{
    HRESULT         ddrval;
    DDSURFACEDESC   ddsd;
    RECT            rc;
    LONG            height,width;

    GetClientRect(hWnd,&rc);
    height = v_size;//rc.bottom;
    width = h_size;//rc.right;

    ddrval = DirectDrawCreate(NULL,&DDraw,NULL);
    if( ddrval != DD_OK )
    {
        MessageBox(hWnd,"error","Direct Draw Create Failed\n",MB_OK);
        return FALSE;
    }

    ddrval = DDraw->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
    if( ddrval != DD_OK )
    {
        MessageBox(hWnd,"error","SetCooperativeLevel Failed\n",MB_OK);
        return FALSE;
    }

    if (BackBuffer) BackBuffer->Release(),BackBuffer=NULL;
    if (FrontBuffer) FrontBuffer->Release(),FrontBuffer=NULL;
    if (Clipper) Clipper->Release(),Clipper=NULL;

    ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |DDSCAPS_VIDEOMEMORY;
    ddrval = DDraw->CreateSurface(&ddsd,&FrontBuffer,NULL);
    if (ddrval != DD_OK)
    {
        ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
        ddrval = DDraw->CreateSurface(&ddsd,&FrontBuffer,NULL);
    }
    if( ddrval != DD_OK )
    {
        MessageBox(hWnd,"error","Create FrontBuffer Failed\n",MB_OK);
        return FALSE;
    }

    ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
    ddsd.dwHeight = height;
    ddsd.dwWidth = width;
    ddrval = DDraw->CreateSurface(&ddsd,&BackBuffer,NULL);
    if (ddrval != DD_OK)
    {
        ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
        ddrval = DDraw->CreateSurface(&ddsd,&BackBuffer,NULL);
    }
    if( ddrval != DD_OK )
    {
        MessageBox(hWnd,"error","Create BackBuffer Failed\n",MB_OK);
        return FALSE;
    }

    ddrval = DDraw->CreateClipper(0,&Clipper,NULL);
    if (ddrval != DD_OK)
    {
        MessageBox(hWnd,"error","Create Clipper Failed\n",MB_OK);
        return FALSE;
    }

    ddrval = Clipper->SetHWnd(0,hWnd);
    if (ddrval != DD_OK)
    {
        MessageBox(hWnd,"error","Set HWnd Failed\n",MB_OK);
        return  FALSE;
    }    

    ddrval = FrontBuffer->SetClipper(Clipper);
    if (ddrval != DD_OK)
    {
        MessageBox(hWnd,"error","Set Clipper Failed\n",MB_OK);
        return FALSE;
    }
    return TRUE;
}

void ReleaseDDraw()
{
    if (BackBuffer) BackBuffer->Release(), BackBuffer =  NULL;
    if (FrontBuffer) FrontBuffer->Release(), FrontBuffer =  NULL;
    if (Clipper) Clipper->Release(), Clipper = NULL;
    if (DDraw) DDraw->Release(),DDraw =  NULL;
}

void DDPutPixel (int x, int y,BYTE r,BYTE g,BYTE b)
{
    int index = x*3+y*NextScan;

    Start[index+2] = b;
    Start[index+1] = g;
    Start[index+0] = r;
}

void Render(void)
{
    HRESULT        ddrval;
    DDBLTFX        ddbltfx;
    DDSURFACEDESC  ddsd;
//    int i;
    if (FrontBuffer->IsLost() == DDERR_SURFACELOST)
      FrontBuffer->Restore();

    ddbltfx.dwSize=sizeof(ddbltfx);
    ddbltfx.dwFillColor=0xffffff;

    BackBuffer->Blt(NULL,NULL,NULL,DDBLT_COLORFILL | DDBLT_WAIT,&ddbltfx);

    ddsd.dwSize=sizeof(ddsd);
    ddsd.dwFlags=DDSD_PITCH | DDSD_ALL;
    ddrval = BackBuffer->Lock(NULL,&ddsd,
                              DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR,NULL);

    if (ddrval != DD_OK)
    {
        return;
    }

    Start = (char*)ddsd.lpSurface;
    NextScan = ddsd.lPitch;

    Image_YCbCr_to_BGR(IMAGE, Start, Ximage, Yimage);

    BackBuffer->Unlock(NULL);

    ddrval = FrontBuffer->Blt(&rcWindow,BackBuffer,NULL,DDBLT_WAIT,NULL);

    if (ddrval != DD_OK)
    {
        return;
    }
}

int WINAPI WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPSTR     lpCmdLine,
                   int       nShowCmd)
{
    //&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&7777
	ZeroMemory(&ofile, sizeof(ofile));
	ofile.lStructSize = sizeof(ofile);
	ofile.hInstance = hInstance;
	ofile.Flags = OFN_FILEMUSTEXIST | OFN_READONLY;
	ofile.lpstrFile = file;
	ofile.lpstrFilter = "All File\0*.*\0";
	ofile.nMaxFile = 256;
	file[0] = NULL;
//&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&77
	/*FILE* in;
	clock_t begin, end;

    char filename[128]="Test1.jpg";*/
    
	UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);
    UNREFERENCED_PARAMETER(nShowCmd);

    //register a window
    char szClassName  [40] = "WinDirectDraw";
    char szWindowTitle[40] = "BI Dozen Mpeg Player";

    WNDCLASSEX wcex;
    wcex.cbSize        = sizeof(wcex);
    wcex.lpszClassName = szClassName;
    wcex.hInstance     = hInstance;
    wcex.hIcon         = LoadIcon(NULL,IDI_APPLICATION);
    wcex.hIconSm       = LoadIcon(NULL,IDI_APPLICATION);
    wcex.hCursor       = LoadCursor(NULL,IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wcex.lpszMenuName  = "mainmenu";
    wcex.cbClsExtra    = 0;
    wcex.cbWndExtra    = 0;
    wcex.style         = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc   = ClientWndProc;
    if (!RegisterClassEx(&wcex)) return FALSE;

    //create main window
    HWND hwnd=NULL;
    hwnd=CreateWindowEx(WS_EX_DLGMODALFRAME,
                        szClassName,
                        szWindowTitle,
                        WS_OVERLAPPEDWINDOW,
                        0,
                        0,
                        352,
                        260,
                        NULL,
                        LoadMenu(hInstance,"mainmenu"),
                        hInstance,
                        NULL);
    if (!hwnd) return FALSE;

    winhwnd=hwnd;
    //if (!InitDDraw(hwnd))
    //return FALSE;

    //show a window
    ShowWindow(hwnd,SW_SHOWNORMAL);
    UpdateWindow(hwnd);

    //message loop
    MSG msg;
    while(GetMessage(&msg,NULL,0,0))
    {
        TranslateMessage(&msg);
        DispatchMessage (&msg);
    }
    return msg.wParam;
}

LRESULT WINAPI ClientWndProc(HWND   hwnd,
                             UINT   msg,
                             WPARAM wParam,
                             LPARAM lParam)
{
    HDC            hdc;
    PAINTSTRUCT    paint;

    switch(msg)
    {
        case WM_CREATE:
        {
            return 0;
        }

        case WM_CLOSE:
        {
            ReleaseDDraw();
            PostQuitMessage(0);
            return 0;
        }

        /*case WM_MOUSEMOVE:
        {
            xPos=LOWORD(lParam);
            yPos=HIWORD(lParam);
            sprintf(buffer2,"x=%d,y=%d",xPos,yPos);
            InvalidateRect(hwnd,NULL,TRUE);

            return 0;
        }*/

        case WM_MOVE:
        {
            GetClientRect(hwnd, &rcWindow);
            ClientToScreen(hwnd, (LPPOINT)&rcWindow);
            ClientToScreen(hwnd, (LPPOINT)&rcWindow+1);
            return 0;
        }

        case WM_COMMAND:
        {
            switch(wParam)
            {

				/*case IDO_TEST:
				{
					CWinApp::OnFileOpen();
				}*/
				case IDO_TEST:
				{
             		GetOpenFileName(&ofile);
					ofile.lpstrFile=file;
					FILE* in;
					in=fopen(file,"rb");
					if(in==NULL){
						printf("File %s does not exist!\n", file);
						return 0;
					}
	                
                    MpegClass mpeg(in);
                    Ximage=mpeg.h_size;
                    Yimage=mpeg.v_size;
                    InitDDraw(hwnd,Ximage,Yimage);
                    int i;
                    for(i=0;;i++){
                        if(mpeg.decode()==0) break;
                        IMAGE=mpeg.picture[mpeg.current_picture];
                        Render();
                    }
                    IMAGE=mpeg.picture[mpeg.future_picture];
                    Render();
                    if(in!=NULL)
						fclose(in);
					return 0;
				}
            }
        }

        case WM_PAINT:
        {
            hdc=BeginPaint(hwnd,&paint);
            if (hdc==NULL) break;

            TextOut(hdc,0,300,buffer2,strlen(buffer2));

            EndPaint(hwnd,&paint);
            return 0;
        }
    }
    return DefWindowProc(hwnd,msg,wParam,lParam);
}
